package ru.dbotthepony.kstarbound

import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonParser
import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.defs.TileDefinitionBuilder
import ru.dbotthepony.kstarbound.defs.TileRenderTemplate
import ru.dbotthepony.kstarbound.world.World
import java.io.File
import java.io.FileNotFoundException
import java.io.FilenameFilter

class TileDefLoadingException(message: String, cause: Throwable? = null) : IllegalStateException(message, cause)

object Starbound {
	private val tiles = HashMap<String, TileDefinition>()
	val tilesAccess = object : Map<String, TileDefinition> by tiles {}

	val world = World()

	private val _filepath = ArrayList<File>()
	val filepath = object : List<File> by _filepath {}

	fun addFilePath(path: File) {
		_filepath.add(path)
	}

	fun findFile(path: File): File {
		if (path.exists()) {
			return path.canonicalFile
		}

		for (sPath in _filepath) {
			val newPath = File(sPath.path, path.path)

			if (newPath.exists()) {
				return newPath
			}
		}

		throw FileNotFoundException("Unable to find $path in any of known file paths")
	}

	fun findFile(path: String) = findFile(File(path))

	fun loadJson(path: String): JsonElement {
		if (path[0] == '/')
			return JsonParser.parseReader(findFile(path.substring(1)).bufferedReader())

		return JsonParser.parseReader(findFile(path).bufferedReader())
	}

	fun loadTileDefinition(name: String): TileDefinition {
		return tiles.computeIfAbsent(name) {
			val foundPath = findFile("tiles/materials/$name.material")
			return@computeIfAbsent TileDefinitionBuilder.fromJson(JsonParser.parseReader(foundPath.bufferedReader()) as JsonObject).build(foundPath.parent)
		}
	}

	fun getTileDefinition(name: String): TileDefinition? {
		return tiles[name]
	}

	fun loadTileMaterials() {
		for (sPath in _filepath) {
			val newPath = File(sPath.path, "tiles/materials")

			if (newPath.exists() && newPath.isDirectory) {
				val findFiles = newPath.listFiles()!!

				for (listedFile in findFiles) {
					if (listedFile.path.endsWith(".material")) {
						try {
							val tileDef = TileDefinitionBuilder.fromJson(JsonParser.parseReader(listedFile.bufferedReader()) as JsonObject).build(listedFile.parent)

							check(tiles[tileDef.materialName] == null) { "Already has material with ID ${tileDef.materialName} loaded!" }
							tiles[tileDef.materialName] = tileDef
						} catch(err: Throwable) {
							throw TileDefLoadingException("Loading tile file ${listedFile.name}", err)
						}
					}
				}
			}
		}
	}
}
