package ru.dbotthepony.kstarbound.math

import com.google.gson.JsonArray
import kotlin.math.cos
import kotlin.math.sin

data class Vector2i(val x: Int, val y: Int) : IMatrixLike, IMatrixLikeGetterI {
	operator fun plus(other: Vector2i) = Vector2i(x + other.x, y + other.y)
	operator fun plus(other: Int) = Vector2i(x + other, y + other)
	operator fun minus(other: Vector2i) = Vector2i(x - other.x, y - other.y)
	operator fun minus(other: Int) = Vector2i(x - other, y - other)
	operator fun times(other: Vector2i) = Vector2i(x * other.x, y * other.y)
	operator fun times(other: Int) = Vector2i(x * other, y * other)
	operator fun div(other: Vector2i) = Vector2i(x / other.x, y / other.y)
	operator fun div(other: Int) = Vector2i(x / other, y / other)

	override val columns = 1
	override val rows = 2

	override fun get(row: Int, column: Int): Int {
		if (column != 0) {
			throw IndexOutOfBoundsException("Column must be 0 ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	companion object {
		fun fromJson(input: JsonArray): Vector2i {
			return Vector2i(input[0].asInt, input[1].asInt)
		}

		val ZERO = Vector2i(0, 0)
	}
}

data class Vector3f(val x: Float = 0f, val y: Float = 0f, val z: Float = 0f) : IMatrixLikeFloat {
	override val columns = 1
	override val rows = 3

	override fun get(row: Int, column: Int): Float {
		if (column != 0) {
			throw IndexOutOfBoundsException("Column must be 0 ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			2 -> z
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	fun rotateAroundThis(rotation: Double): Matrix4f {
		val c = cos(rotation).toFloat()
		val s = sin(rotation).toFloat()
		val cInv = 1f - c

		return Matrix4f(
			m00 = c + x * x * cInv,			m01 = x * y * cInv - z * s,		m02 = x * z * cInv + y * s,
			m10 = y * x * cInv + z * s,		m11 = c + y * y * cInv,			m12 = y * z * cInv - x * s,
			m20 = z * x * cInv - y * s,		m21 = z * y * cInv + x * s,		m22 = c + z * z * cInv,
		)
	}

	operator fun times(other: IMatrixLikeFloat): Vector3f {
		if (other.rows >= 4 && other.columns >= 4) {
			val x = this.x * other[0, 0] +
					this.y * other[0, 1] +
					this.z * other[0, 2] +
					other[0, 3]

			val y = this.x * other[1, 0] +
					this.y * other[1, 1] +
					this.z * other[1, 2] +
					other[1, 3]

			val z = this.x * other[2, 0] +
					this.y * other[2, 1] +
					this.z * other[2, 2] +
					other[2, 3]

			return Vector3f(x, y, z)
		} else if (other.rows >= 3 && other.columns >= 3) {
			val x = this.x * other[0, 0] +
					this.y * other[0, 1] +
					this.z * other[0, 2]

			val y = this.x * other[1, 0] +
					this.y * other[1, 1] +
					this.z * other[1, 2]

			val z = this.x * other[2, 0] +
					this.y * other[2, 1] +
					this.z * other[2, 2]

			return Vector3f(x, y, z)
		}

		throw IllegalArgumentException("Incompatible matrix provided: ${other.rows} x ${other.columns}")
	}

	companion object {
		val UP = Vector3f(0f, 1f, 0f)
		val DOWN = Vector3f(0f, -1f, 0f)
		val LEFT = Vector3f(-1f, 0f, 0f)
		val RIGHT = Vector3f(1f, 0f, 0f)
		val FORWARD = Vector3f(0f, 0f, 1f)
		val BACKWARD = Vector3f(0f, 0f, -1f)
	}
}

data class Vector4f(val x: Float, val y: Float, val z: Float, val w: Float) : IMatrixLikeFloat {
	override val columns = 1
	override val rows = 4

	override fun get(row: Int, column: Int): Float {
		if (column != 0) {
			throw IndexOutOfBoundsException("Column must be 0 ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			2 -> z
			3 -> w
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	operator fun times(other: IMatrixLikeFloat): Vector4f {
		if (other.rows >= 4 && other.columns >= 4) {
			val x = this.x * other[0, 0] +
					this.y * other[0, 1] +
					this.z * other[0, 2] +
					this.w * other[0, 3]

			val y = this.x * other[1, 0] +
					this.y * other[1, 1] +
					this.z * other[1, 2] +
					this.w * other[1, 3]

			val z = this.x * other[2, 0] +
					this.y * other[2, 1] +
					this.z * other[2, 2] +
					this.w * other[2, 3]

			val w = this.x * other[3, 0] +
					this.y * other[3, 1] +
					this.z * other[3, 2] +
					this.w * other[3, 3]

			return Vector4f(x, y, z, w)
		}

		throw IllegalArgumentException("Incompatible matrix provided: ${other.rows} x ${other.columns}")
	}
}
